/********************************************************************/
/*                                                                  */
/*                   AutoMap/AutoLink Users' Manual                 */
/*                                                                  */
/*      Example of sending a double-linked list with AutoLink       */
/*                                                                  */
/* Creation date............April 21 2000                           */
/* Last modification date...May   18 2000                           */
/*                                                                  */
/* Author : Olivier NICOLAS (olivier.nicolas@nist.gov)              */
/*          Guest researcher at NIST                                */
/*                                                                  */
/********************************************************************/
#include "perf.h"
#include "mpi.h"
#include "autolink.inc"
#include "randomizer.h"
#include "timer.h"
#include "myprint.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


short Integrity;
elementT *list;
elementT *ptr_list;


void createData()
{
  int i;

  myPrintf("Rank[0]... Creating Data\n");

  initRandomizer();
  initTimer();

  list = malloc(sizeof(elementT));
  list->value = rand_R(1000);
  strcpy(list->name, "first");
  
  ptr_list = list;
  
  for (i=1; i<LIST_LENGHT; i++)
    {
      elementT *temp;
      
      temp = malloc(sizeof(elementT));
      temp->value = rand_R(1000);
      sprintf(temp->name , "%d", temp->value);
      
      ptr_list->next = temp;
      ptr_list = temp;
    }
  ptr_list->next = NULL;

  myPrintf("Rank[0]... Data created!\n");
}

void freeData(int rank)
{
  int i;
  elementT *tmp;

  myPrintf("Rank[%d]... Freeing Data\n", rank);

  ptr_list = list;

  for(i=0; i<LIST_LENGHT; i++)
    {
      tmp=ptr_list;
      ptr_list=ptr_list->next;
      free(tmp);
    }
}

void printData(int rank)
{
  int i;
  ptr_list = list;

  if (Integrity == 1)
    {
      printf("Rank[%d]... Printing Data\n", rank);
      
      for(i=0; i<LIST_LENGHT; i++)
	{
	  printf("[rank %d] (%d %s)\n", rank, ptr_list->value, ptr_list->name);
	  ptr_list=ptr_list->next;
	}
    }
}

void sendData(int rank, int sendTo)
{
  myPrintf("Rank[%d]... Sending Data to proc%d\n", rank, sendTo);

  if (AL_Send(list, AL_elementT, sendTo, 0, MPI_COMM_WORLD)!= MPI_SUCCESS)
    {
      printf("Error in Sending");
      exit(1);
    }

  myPrintf("Rank[%d]... Data Sent\n", rank);
}

void recvData(int rank, int recvFrom)
{
  MPI_Status status;
  int type;

  myPrintf("Rank[%d]... Receiving Data from proc%d\n", rank, recvFrom);
  
  if (AL_Recv((void *) &list, &type, recvFrom, 0, MPI_COMM_WORLD, &status, NULL) != MPI_SUCCESS) 
    {
      printf("Error in Receiving");
      exit(1);
    }

  myPrintf("Rank[%d]... Data Received\n", rank);
}


int main(int argc, char* argv[])
{
  int my_rank;
  int num_ranks;
  int recvFrom, sendTo;

  MPI_Init(&argc ,&argv);
  AL_Init(-1);

  MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);    /* get rank */
  MPI_Comm_size(MPI_COMM_WORLD, &num_ranks);  /* get number of ranks */
  
  if (num_ranks == 1)
    {
      if (my_rank == 0)
	{
	  printf("Use a number of process >= 2\n");
	}
    }
  else
    {
      recvFrom = my_rank-1;
      sendTo = (my_rank+1)%num_ranks;
      
      if (recvFrom<0)
	recvFrom = num_ranks-1;
      
      while (--argc>0 && (*++argv)[0] == '-')
	{
	  if (strcmp("-integrity", *argv) == 0 )
	    Integrity = 1;
	  else if (strcmp("-detail", *argv) == 0 )
	    Detail = 1;
	}
      
      if (my_rank == 0) 
	{
	  createData();
	  printData(my_rank);
	  beginTimer();
	  sendData(my_rank, sendTo);
	  recvData(my_rank, recvFrom);
	  printf("%d ways time : %dms\n", num_ranks, endTimer());
	} 
      else 
	{
	  recvData(my_rank, recvFrom);
	  sendData(my_rank, sendTo);
	  printData(my_rank);
	} 
      
      freeData(my_rank);
      myPrintf("Rank[%d]... Finalizing\n", my_rank);
    }

  MPI_Finalize();
  AL_Finalize();
  
  return (0);
}
