/********************************************************************/
/*                                                                  */
/*                   AutoMap/AutoLink Users' Manual                 */
/*                                                                  */
/*                        A concrete example                        */
/*                                                                  */
/*             From a Generic Genetic Programming Project           */
/*                                                                  */
/*       Sending of the initial population of genetic programs      */
/*                          with AutoLink                           */
/*                                                                  */
/* Creation date............April 21 2000                           */
/* Last modification date...May   18 2000                           */
/*                                                                  */
/* Author : Olivier NICOLAS (olivier.nicolas@nist.gov)              */
/*          Guest researcher at NIST                                */
/*                                                                  */
/********************************************************************/
/*                                                                  */
/*         Creation of the population and transfer functions        */
/*                                                                  */
/********************************************************************/
#include "perf.h"
#include "randomizer.h"
#include "myprint.h"
#include "individual.h"
#include "randomizer.h"
#include "node.h"
#include "macros.h"
#include "population.h"
#include "timer.h"

#include "mpi.h"
#include "autolink.inc"

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/dir.h>

short Integrity;

void createData()
{
  myPrintf("Rank[0]... Creating Data\n");

  initRandomizer();
  initTimer();

  createInitialPopulation();

  myPrintf("Rank[0]... Data created!\n");
}

void freeData(int rank)
{
  myPrintf("Rank[%d]... Freeing Data\n", rank);

  freePopulation();
}

void printData(int rank)
{
  int i;

  if (Integrity == 1)
    {
      printf("Rank[%d]... Printing Data [individual0]\n", rank);
      
      printf("------------------------------\n");
      printProg(stdout, INDIV(0));
      printf("------------------------------\n");
    }
}

void sendData(int rank, int sendTo)
{
  int i;

  myPrintf("Rank[%d]... Sending Data [All individuals] to proc%d\n", rank, sendTo);

  for (i=0; i<POPULATION_SIZE; i++)
    {
      if (AL_Send(INDIV(i), AL_individualT, sendTo, 0, MPI_COMM_WORLD)!= MPI_SUCCESS)
	{
	  printf("Error in Sending");
	  exit(1);
	}
    }

  myPrintf("Rank[%d]... Data Sent\n", rank);
}

void recvData(int rank, int recvFrom)
{
  int i;
  MPI_Status status;
  int type;

  myPrintf("Rank[%d]... Receiving Data from proc%d\n", rank, recvFrom);
  
  for (i=0; i<POPULATION_SIZE; i++)
    {
      if (AL_Recv((void *) &population.individuals[i], &type, recvFrom, 0, MPI_COMM_WORLD, &status, NULL) != MPI_SUCCESS) 
	{
	  printf("Error in Receiving");
              exit(1);
	}
    }
  
  myPrintf("Rank[%d]... Data Received\n", rank);
}


int main(int argc, char* argv[])
{
  int my_rank;
  int num_ranks;
  int recvFrom, sendTo;

  MPI_Init(&argc ,&argv);
  AL_Init(-1);

  MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);    /* get rank */
  MPI_Comm_size(MPI_COMM_WORLD, &num_ranks);  /* get number of ranks */

  if (num_ranks == 1)
    {
      if (my_rank == 0)
	{
	  printf("Use a number of process >= 2\n");
	}
    }
  else
    {
      recvFrom = my_rank-1;
      sendTo = (my_rank+1)%num_ranks;

      if (recvFrom<0)
	recvFrom = num_ranks-1;

      while (--argc>0 && (*++argv)[0] == '-')
	{
	  if (strcmp("-integrity", *argv) == 0 )
	    Integrity = 1;
	  else if (strcmp("-detail", *argv) == 0 )
	    Detail = 1;
	}

      if (my_rank == 0) 
	{
	  createData();
	  printData(my_rank);
	  beginTimer();
	  sendData(my_rank, sendTo);
	  recvData(my_rank, recvFrom);
	  printf("%d ways time : %dms\n", num_ranks, endTimer());
	} 
      else 
	{
	  recvData(my_rank, recvFrom);
	  sendData(my_rank, sendTo);
	  printData(my_rank);
	} 
  
      freeData(my_rank);
      myPrintf("Rank[%d]... Finalizing\n", my_rank);
    }
  
  MPI_Finalize();
  AL_Finalize();

  return (0);
}
