/********************************************************************/
/*                                                                  */
/*                   AutoMap/AutoLink Users' Manual                 */
/*                                                                  */
/*                        A concrete example                        */
/*                                                                  */
/*             From a Generic Genetic Programming Project           */
/*                                                                  */
/*       Sending of the initial population of genetic programs      */
/*                          with AutoLink                           */
/*                                                                  */
/* Creation date............April 21 2000                           */
/* Last modification date...May   18 2000                           */
/*                                                                  */
/* Author : Olivier NICOLAS (olivier.nicolas@nist.gov)              */
/*          Guest researcher at NIST                                */
/*                                                                  */
/********************************************************************/
/*                                                                  */
/*                 Sub-trees management functions                   */
/*                                                                  */
/********************************************************************/
#include "node.h"
#include "spec.h"
#include "randomizer.h"
#include "individual.h"

#include <stdlib.h>
#include <stdio.h>

int rank(nodeT *node, int subTreeNb)
{
  int i;

  for (i=0; i<dimensions[subTreeNb]; i++)
    if (Trees[subTreeNb].Nodes[i].id == node->op)
      return i;

  printf("Unreachable point %d %d\n", node->op, subTreeNb);

  return (-1);
}

nodeT *produceBranch(int depth, int subTreeNb)
{
  int i;
  nodeT *node;
  int kindOfNode;
  unsigned short functionOrTerminal = rand_R(1);/* 0 for function, 1 for terminal */

  /* chose a kind of node */
  kindOfNode = rand_R(dimensions[subTreeNb]-1);

  if (depth == 0) /* creates a leaf, in order to respect the depth limitation */
    functionOrTerminal = 1;

  switch(functionOrTerminal)
    {
    case 0:
      while (Trees[subTreeNb].Nodes[kindOfNode].arity==LEAF)
	kindOfNode = rand_R(dimensions[subTreeNb]-1);
      break;
    case 1:
      while (Trees[subTreeNb].Nodes[kindOfNode].arity!=LEAF)
	kindOfNode = rand_R(dimensions[subTreeNb]-1);
      break;
    }

  node = createNode(Trees[subTreeNb].Nodes[kindOfNode].id);

  for (i=0; i<ARITY(node, subTreeNb); i++)
    ARGUMENT(node, i) = produceBranch(depth-1, subTreeNb);

  for (i=0, node->entryPoints = 1; i<ARITY(node, subTreeNb); i++)
    node->entryPoints+=ARGUMENT(node, i)->entryPoints;

  node->depth = depth;

  return node;
}

nodeT *createNode(int op) 
{
  nodeT *node = NULL;
  int i;

  if ( (node = malloc(sizeof(nodeT)) ) != NULL) 
    {
      if( op == CONST) /* randomly generate a "RbiggerReal" */
	node->value = rand_R(10000);

      node->op   = op;
      
      for(i=0; i < MAX_ARG; i++)
	ARGUMENT(node, i) = NULL;
    }
  else 
    {
      fprintf(stderr, "[ERROR 2] createNode : Not enough memory for node allocation\n");
    }

  return(node);
}

void freeNode(nodeT *node, int subTreeNb) 
{
  /* recursively frees the memory of son-nodes and then the current one */
  int i;

 for(i=0; i<ARITY(node, subTreeNb); i++)
    freeNode(ARGUMENT(node, i), subTreeNb);

 free(node);
}


void printTree(FILE *stream, nodeT *node, int subTreeNb)
{
  int i;

  switch (ARITY(node, subTreeNb))
    {
    case LEAF :
      if (node->op == CONST)
	{
	  fprintf(stream, "%d", node->value);fflush(stream);
	}
      else
	{
	  fprintf(stream, "%s", NAME(node, subTreeNb));fflush(stream);
	}
      break;

    default :
        fprintf(stream, "%s(", NAME(node, subTreeNb));fflush(stream);

      for(i=0; i<ARITY(node,subTreeNb)-1; i++)
	{
	  printTree(stream, ARGUMENT(node, i), subTreeNb);
	  fprintf(stream, ", ");fflush(stream);
	}

      printTree(stream, ARGUMENT(node, i), subTreeNb);
      fprintf(stream, ")");fflush(stream);

      break;
    }
}
