/********************************************************************/
/*                                                                  */
/*                  AutoMap/AutoLink Users' Manual                  */
/*                                                                  */
/*          Example on sending an array of "struct element"         */
/*                           version 1                              */
/*                                                                  */
/* Creation date............April 21 2000                           */
/* Last modification date...May   18 2000                           */
/*                                                                  */
/* Author : Olivier NICOLAS (olivier.nicolas@nist.gov)              */
/*          Guest researcher at NIST                                */
/*                                                                  */
/********************************************************************/
#include "array.h"
#include "mpi.h"
#include "mpitypes.inc"
#include "randomizer.h"
#include "myprint.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


element myArray[TEST_ARRAY_SIZE];


void createData()
{
  int i;

  myPrintf("Rank[0] ... Creating Data\n");

  initRandomizer();

  for (i=0; i<TEST_ARRAY_SIZE; i++)
    {
      myArray[i].value = rand_R(10000);
      sprintf(myArray[i].name, "S%d", i);
      myArray[i].value2 = (float) rand_R(100);
    }
}

void freeData(int rank)
{
  myPrintf("Rank[%d] ... Freeing Data\n", rank);

  /* nothing to do here */
}

void printData(int rank)
{
  int i;

  myPrintf("Rank[%d] ... Printing Data\n", rank);

  printf("Rank[%d]---> ", rank);
  for(i=0; i<TEST_ARRAY_SIZE; i++)
   printf("%d %s %.2f ", myArray[i].value, myArray[i].name, myArray[i].value2);
  printf("\n");
}

void sendData(int rank, int sendTo)
{
  myPrintf("Rank[%d] ... Sending Data\n", rank);

  if (MPI_Send(&myArray, TEST_ARRAY_SIZE, AM_element, sendTo, 0, MPI_COMM_WORLD)!= MPI_SUCCESS)
    {
      printf("Error in Sending");
      exit(1);
    }

  myPrintf("Rank[%d] ... Data Sent\n", rank);
}

void recvData(int rank, int recvFrom)
{
  MPI_Status status;

  myPrintf("Rank[%d] ... Receiving Data\n", rank);
  
  if (MPI_Recv(myArray, TEST_ARRAY_SIZE, AM_element, recvFrom, 0, MPI_COMM_WORLD, &status) != MPI_SUCCESS) 
    {
      printf("Error in Receiving");
      exit(1);
    }
  
  myPrintf("Rank[%d] ... Data Received\n", rank);
}


int main(int argc,char* argv[])
{
  int my_rank;
  int num_ranks;

  MPI_Init(&argc ,&argv);
  Build_MPI_Types();

  MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);    /* get rank */
  MPI_Comm_size(MPI_COMM_WORLD, &num_ranks);  /* get number of ranks */
    
  while (--argc>0 && (*++argv)[0] == '-')
    {
      if (strcmp("-detail", *argv) == 0 )
        Detail = 1;
    }

  if (my_rank == 0) 
    {
      int sendTo = 1;

      createData();
      printData(my_rank);
      sendData(my_rank, sendTo);
      freeData(my_rank);
    } 
  else 
    {
      int recvFrom = 0;

      recvData(my_rank, recvFrom);
      printData(my_rank);
      freeData(my_rank);
    } 
  
  myPrintf("Rank[%d] ... Finalizing\n", my_rank);
  MPI_Finalize();

  return (0);
}
