/********************************************************************/
/*                                                                  */
/*                   AutoMap/AutoLink Users' Manual                 */
/*                                                                  */
/*               Example on sending an array of arrays              */
/*                                                                  */
/* Creation date............April 21 2000                           */
/* Last modification date...May   18 2000                           */
/*                                                                  */
/* Author : Olivier NICOLAS (olivier.nicolas@nist.gov)              */
/*          Guest researcher at NIST                                */
/*                                                                  */
/********************************************************************/
#include "arrayOfArrays.h"
#include "mpi.h"
#include "mpitypes.inc"
#include "randomizer.h"
#include "myprint.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


pArrayT *myArray;


void createData()
{
  int i, j;

  myPrintf("Rank[0] ... Creating Data\n");

  initRandomizer();

  myArray = malloc(sizeof(pArrayT));
  
  for (i=0; i<ARRAY_SIZE; i++)
    {
      for (j=0; j<INT_ARRAY_SIZE; j++)
	myArray->array[i].intValues[j] = rand_R(100);
      
      sprintf(myArray->array[i].name, "Val %d", i+1);
      
      for (j=0; j<FLOAT_ARRAY_SIZE; j++)
	myArray->array[i].floatValues[j] = rand_R(10000)/3.12;
    }
}

void freeData(int rank)
{
  myPrintf("Rank[%d] ... Freeing Data\n", rank);

  free(myArray);
}

void printData(int rank)
{
  int i, j;

  printf("Rank %d ---------------------\n", rank);
  for(i=0; i<ARRAY_SIZE; i++)
    {
      printf("\t[%d] (%s |", i, myArray->array[i].name);

      for (j=0; j<INT_ARRAY_SIZE; j++)
	printf(" %d", myArray->array[i].intValues[j]);
      
      printf(" |");

      for (j=0; j<FLOAT_ARRAY_SIZE; j++)
	printf(" %.3f", myArray->array[i].floatValues[j]);

      printf(")\n");
    }

  printf("----------------------------\n");
}

void sendData(int rank, int sendTo)
{
  myPrintf("Rank[%d] ... Sending Data\n", rank);

  if (MPI_Send(myArray, 1, AM_pArrayT, sendTo, 0, MPI_COMM_WORLD)!= MPI_SUCCESS)
    {
      printf("Error in Sending");
      exit(1);
    }
  
  myPrintf("Rank[%d] ... Data Sent\n", rank);
}

void recvData(int rank, int recvFrom)
{
  MPI_Status status;

  myArray = malloc(sizeof(pArrayT));

  myPrintf("Rank[%d] ... Receiving Data\n", rank);
  
  if (MPI_Recv(myArray, 1, AM_pArrayT, recvFrom, 0, MPI_COMM_WORLD, &status) != MPI_SUCCESS) 
    {
      printf("Error in Receiving");
      exit(1);
    }
  
  myPrintf("Rank[%d] ... Data Received\n", rank);
}


int main(int argc,char* argv[])
{
  int my_rank;
  int num_ranks;

  MPI_Init(&argc ,&argv);
  Build_MPI_Types();

  MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);    /* get rank */
  MPI_Comm_size(MPI_COMM_WORLD, &num_ranks);  /* get number of ranks */
    
  while (--argc>0 && (*++argv)[0] == '-')
    {
      if (strcmp("-detail", *argv) == 0 )
        Detail = 1;
    }

  if (my_rank == 0) 
    {
      int sendTo = 1;

      createData();
      printData(my_rank);
      sendData(my_rank, sendTo);
      freeData(my_rank);
    } 
  else 
    {
      int recvFrom = 0;

      recvData(my_rank, recvFrom);
      printData(my_rank);
      freeData(my_rank);
    } 
  
  myPrintf("Rank[%d] ... Finalizing\n", my_rank);
  MPI_Finalize();

  return (0);
}
