/********************************************************************/
/*                                                                  */
/*                  AutoMap/AutoLink Users' Manual                  */
/*                                                                  */
/*   This program is for you to check if your MPI version suffers   */
/*                    from the padding problem                      */
/*                                                                  */
/* Creation date............April 21 2000                           */
/* Last modification date...May   18 2000                           */
/*                                                                  */
/* Author : Olivier NICOLAS (olivier.nicolas@nist.gov)              */
/*          Guest researcher at NIST                                */
/*                                                                  */
/********************************************************************/
#include "mpi.h"
#include "padding.h"
#include "padtypes.inc"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

pArray1T myPopulation1;
pArray2T myPopulation2;

void sendData(int rank, int sendTo)
{
  if (MPI_Send(&myPopulation1, 1, AM_pArray1T, sendTo, 0, MPI_COMM_WORLD)!= MPI_SUCCESS)
    {
      printf("Error in Sending");
      exit(1);
    }
  if (MPI_Send(&myPopulation2, 1, AM_pArray2T, sendTo, 0, MPI_COMM_WORLD)!= MPI_SUCCESS)
    {
      printf("Error in Sending");
      exit(1);
    }
}

void recvData(int rank, int recvFrom)
{
  MPI_Status status;
  
  if (MPI_Recv(&myPopulation1, 1, AM_pArray1T, recvFrom, 0, MPI_COMM_WORLD, &status) != MPI_SUCCESS) 
    {
      printf("Error in Receiving");
      exit(1);
    }
  
  if (MPI_Recv(&myPopulation2, 1, AM_pArray2T, recvFrom, 0, MPI_COMM_WORLD, &status) != MPI_SUCCESS) 
    {
      printf("Error in Receiving");
      exit(1);
    }
}


int main(int argc,char* argv[])
{
  int my_rank;
  int num_ranks;

  MPI_Init(&argc ,&argv);
  Build_MPI_Types();

  MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);    /* get rank */
  MPI_Comm_size(MPI_COMM_WORLD, &num_ranks);  /* get number of ranks */
  
  if (my_rank == 0) 
    {
      int sendTo = 1;

      strcpy(myPopulation1.population[1].name, "ch12");
      strcpy(myPopulation2.population[1].name, "ch22");

      sendData(my_rank, sendTo);
    } 
  else 
    {
      int recvFrom = 0;

      recvData(my_rank, recvFrom);

      if (strcmp(myPopulation2.population[1].name, "ch22") == 0)
	{
	  if (strcmp(myPopulation1.population[1].name, "ch12") == 0)
	    printf("padding compliant\n");
	  else
	    printf("padding problem present\n");
	}
      else
	printf("padding problem persistent\n");
    } 
  
  MPI_Finalize();

  return (0);
}
