/********************************************************************/
/*                                                                  */
/*                  AutoMap/AutoLink Users' Manual                  */
/*                                                                  */
/*       This example uses the MPI function MPI_Reduce; this        */
/*  implementation is without AutoMap. For a comparison, you also   */
/*  have an implementation with AutoMap (directory ../withAutoMap)  */
/*                                                                  */
/* Creation date............April 21 2000                           */
/* Last modification date...May   18 2000                           */
/*                                                                  */
/* Author : Olivier NICOLAS (olivier.nicolas@nist.gov)              */
/*          Guest researcher at NIST                                */
/*                                                                  */
/* This program originally from Thibault Rouffineau                 */
/*                                                                  */
/********************************************************************/
#include <mpi.h>
#include "reduce.h"
#include <stdio.h>
#define true  1
#define false 0

void complexcompare(Complex * in, Complex * inout, int * len, MPI_Datatype * dptr)
{
  int i;

  for (i = 0; i < *len; i++) {
    if (((in->real*in->real) + (in->imag*in->imag))  >
	((inout->real*inout->real)+(inout->imag*inout->imag)))
      *inout = *in;    
    in++;
    inout++;
  }
}

main(int argc, char* argv[])
{
  int my_rank;
  int num_ranks;
  MPI_Op myOp;
  Complex a[3], answer[3];
  MPI_Datatype AM_Complex;

  MPI_Init(&argc, &argv); 

  MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);    /* get rank */
  MPI_Comm_size(MPI_COMM_WORLD, &num_ranks);  /* get number of ranks */

  MPI_Type_contiguous(2, MPI_DOUBLE, &AM_Complex);
  MPI_Type_commit(&AM_Complex);
  
  MPI_Op_create((MPI_User_function *) complexcompare, 0, &myOp);

  if(my_rank == 0) 
    {
      a[0].real = 4.5;
      a[0].imag = 0.0;
      a[1].real = -5;
      a[1].imag = 0.1;
      a[2].real = 6.0;
      a[2].imag = -5.1;
    }
  else if(my_rank == 1) 
    {
      a[0].real = 2.1;
      a[0].imag = 9.0;
      a[1].real = -3.84;
      a[1].imag = 3.0;
      a[2].real = 0.2;
      a[2].imag = 0.8;
    }
  else 
    {
      a[0].real = -1.5;
      a[0].imag = 0.4;
      a[1].real = 7.3;
      a[1].imag = 8.1;
      a[2].real = 2.4;
      a[2].imag = -1.0;
    }

  MPI_Reduce (a, answer, 3, AM_Complex , myOp, 0, MPI_COMM_WORLD);

  if(my_rank == 0)
    printf("Rank 0 Reduce Results\n\n\t[(%.1f,%.1f) (%.1f,%.1f) (%.1f,%.1f)]\n", answer[0].real, answer[0].imag , answer[1].real, answer[1].imag ,answer[2].real, answer[2].imag);

  MPI_Finalize();
  /* Finalizes MPI */
}
